@@ -48,18 +48,18 @@ class MultilineText {
 public:
     MultilineText(const QFont &font, const QString &text) {
         QStringList lines = text.split("\n");
-        int currentY = 0;
+        int width = 0;
+        QFontMetrics fontMetrics(font);
+        fontHeight_ = fontMetrics.ascent() + fontMetrics.descent();
         for (const auto &line : lines) {
             layouts_.emplace_back(std::make_unique<QTextLayout>(line));
             layouts_.back()->setFont(font);
             doLayout(*layouts_.back());
-            auto size = layouts_.back()->boundingRect().toRect();
-            boundingRect_ =
-                boundingRect_.united(size.translated(QPoint(0, currentY)));
-            currentY += size.height();
+            width = std::max(width,
+                             layouts_.back()->boundingRect().toRect().width());
         }
-        QFontMetrics fontMetrics(font);
-        fontHeight_ = fontMetrics.ascent() + fontMetrics.descent();
+        boundingRect_.setTopLeft(QPoint(0, 0));
+        boundingRect_.setSize(QSize(width, lines.size() * fontHeight_));
     }
 
     bool isEmpty() const { return layouts_.empty(); }
@@ -70,8 +70,7 @@ public:
         int currentY = 0;
         for (const auto &layout : layouts_) {
             layout->draw(painter, position + QPoint(0, currentY));
-            auto size = layout->boundingRect();
-            currentY += size.height();
+            currentY += fontHeight_;
         }
         painter->restore();
     }
@@ -84,12 +83,21 @@ private:
     QRect boundingRect_;
 };
 
-FcitxCandidateWindow::FcitxCandidateWindow(FcitxQtICData *data,
-                                           FcitxTheme *theme)
-    : QWindow(), theme_(theme), parent_(data->window()) {
-    setFlags(Qt::ToolTip | Qt::FramelessWindowHint |
-             Qt::BypassWindowManagerHint | Qt::WindowDoesNotAcceptFocus |
-             Qt::NoDropShadowWindowHint);
+FcitxCandidateWindow::FcitxCandidateWindow(QWindow *window,
+                                           QFcitxPlatformInputContext *context)
+    : QWindow(), context_(context), theme_(context->theme()), parent_(window) {
+    constexpr Qt::WindowFlags commonFlags = Qt::FramelessWindowHint |
+                                            Qt::WindowDoesNotAcceptFocus |
+                                            Qt::NoDropShadowWindowHint;
+    if (isWayland_) {
+        // Qt::ToolTip ensures wayland doesn't grab focus.
+        // Not using Qt::BypassWindowManagerHint ensures wayland handle
+        // fractional scale.
+        setFlags(Qt::ToolTip | commonFlags);
+    } else {
+        // Qt::Popup ensures X11 doesn't apply tooltip animation under kwin.
+        setFlags(Qt::Popup | Qt::BypassWindowManagerHint | commonFlags);
+    }
     if (isWayland_) {
         setTransientParent(parent_);
     }
@@ -97,18 +105,7 @@ FcitxCandidateWindow::FcitxCandidateWindow(FcitxQtICData *data,
     surfaceFormat.setAlphaBufferSize(8);
     setFormat(surfaceFormat);
     backingStore_ = new QBackingStore(this);
-    connect(data->window(), &QWindow::visibilityChanged, this,
-            [data](bool visible) {
-                if (!visible) {
-                    data->resetCandidateWindow();
-                }
-            });
-    connect(data->watcher(), &FcitxQtWatcher::availabilityChanged, this,
-            [data](bool avail) {
-                if (!avail) {
-                    data->resetCandidateWindow();
-                }
-            });
+    connect(this, &QWindow::visibleChanged, this, [this] { hoverIndex_ = -1; });
 }
 
 FcitxCandidateWindow::~FcitxCandidateWindow() {}
@@ -204,9 +201,8 @@ void FcitxCandidateWindow::render(QPainter *painter) {
     if (!upperLayout_.text().isEmpty()) {
         upperLayout_.draw(
             painter, topLeft + QPoint(textMargin.left(), textMargin.top()));
-        auto size = upperLayout_.boundingRect();
         // Draw cursor
-        currentHeight += std::max(minH, qCeil(size.height())) + extraH;
+        currentHeight += minH + extraH;
         if (cursor_ >= 0) {
             auto line = upperLayout_.lineForTextPosition(cursor_);
             if (line.isValid()) {
@@ -228,8 +224,7 @@ void FcitxCandidateWindow::render(QPainter *painter) {
         lowerLayout_.draw(painter,
                           topLeft + QPoint(textMargin.left(),
                                            textMargin.top() + currentHeight));
-        auto size = upperLayout_.boundingRect();
-        currentHeight += std::max(minH, qCeil(size.height())) + extraH;
+        currentHeight += minH + extraH;
     }
 
     bool vertical = theme_->vertical();
@@ -321,12 +316,12 @@ void FcitxCandidateWindow::render(QPainter *painter) {
     }
 }
 
-void UpdateLayout(QTextLayout &layout, const QFont &font,
+void UpdateLayout(QTextLayout &layout, const FcitxTheme &theme,
                   std::initializer_list<
                       std::reference_wrapper<const FcitxQtFormattedPreeditList>>
                       texts) {
     layout.clearFormats();
-    layout.setFont(font);
+    layout.setFont(theme.font());
     QVector<QTextLayout::FormatRange> formats;
     QString str;
     int pos = 0;
@@ -347,13 +342,8 @@ void UpdateLayout(QTextLayout &layout, const QFont &font,
                 format.setFontItalic(true);
             }
             if (preedit.format() & FcitxTextFormatFlag_HighLight) {
-                QBrush brush;
-                QPalette palette;
-                palette = QGuiApplication::palette();
-                format.setBackground(QBrush(QColor(
-                    palette.color(QPalette::Active, QPalette::Highlight))));
-                format.setForeground(QBrush(QColor(palette.color(
-                    QPalette::Active, QPalette::HighlightedText))));
+                format.setBackground(theme.highlightBackgroundColor());
+                format.setForeground(theme.highlightColor());
             }
             formats.append(QTextLayout::FormatRange{
                 pos, static_cast<int>(preedit.string().length()), format});
@@ -364,27 +354,25 @@ void UpdateLayout(QTextLayout &layout, const QFont &font,
     layout.setFormats(formats);
 }
 
-void fcitx::FcitxCandidateWindow::updateClientSideUI(
-    const fcitx::FcitxQtFormattedPreeditList &preedit, int cursorpos,
-    const fcitx::FcitxQtFormattedPreeditList &auxUp,
-    const fcitx::FcitxQtFormattedPreeditList &auxDown,
-    const fcitx::FcitxQtStringKeyValueList &candidates, int candidateIndex,
+void FcitxCandidateWindow::updateClientSideUI(
+    const FcitxQtFormattedPreeditList &preedit, int cursorpos,
+    const FcitxQtFormattedPreeditList &auxUp,
+    const FcitxQtFormattedPreeditList &auxDown,
+    const FcitxQtStringKeyValueList &candidates, int candidateIndex,
     int layoutHint, bool hasPrev, bool hasNext) {
-    bool preeditVisble = (cursorpos >= 0 || !preedit.isEmpty());
+    bool preeditVisible = !preedit.isEmpty();
     bool auxUpVisbile = !auxUp.isEmpty();
     bool auxDownVisible = !auxDown.isEmpty();
     bool candidatesVisible = !candidates.isEmpty();
     bool visible =
-        preeditVisble || auxUpVisbile || auxDownVisible || candidatesVisible;
-    auto window = QGuiApplication::focusWindow();
-    if (!theme_ || !visible || !QGuiApplication::focusWindow() ||
-        window != parent_) {
+        preeditVisible || auxUpVisbile || auxDownVisible || candidatesVisible;
+    auto window = context_->focusWindowWrapper();
+    if (!theme_ || !visible || !window || window != parent_) {
         hide();
-        hoverIndex_ = -1;
         return;
     }
 
-    UpdateLayout(upperLayout_, theme_->font(), {auxUp, preedit});
+    UpdateLayout(upperLayout_, *theme_, {auxUp, preedit});
     if (cursorpos >= 0) {
         int auxUpLength = 0;
         for (const auto &auxUpText : auxUp) {
@@ -398,7 +386,7 @@ void fcitx::FcitxCandidateWindow::updateClientSideUI(
         cursor_ = -1;
     }
     doLayout(upperLayout_);
-    UpdateLayout(lowerLayout_, theme_->font(), {auxDown});
+    UpdateLayout(lowerLayout_, *theme_, {auxDown});
     doLayout(lowerLayout_);
     labelLayouts_.clear();
     candidateLayouts_.clear();
@@ -415,18 +403,36 @@ void fcitx::FcitxCandidateWindow::updateClientSideUI(
 
     actualSize_ = sizeHint();
 
-    QRect cursorRect =
-        QGuiApplication::inputMethod()->cursorRectangle().toRect();
+    if (actualSize_.width() <= 0 || actualSize_.height() <= 0) {
+        hide();
+        return;
+    }
+
+#if QT_VERSION >= QT_VERSION_CHECK(5, 14, 0)
+    QSize sizeWithoutShadow = actualSize_.shrunkBy(theme_->shadowMargin());
+#else
+    QSize sizeWithoutShadow =
+        actualSize_ -
+        QSize(theme_->shadowMargin().left() + theme_->shadowMargin().right(),
+              theme_->shadowMargin().top() + theme_->shadowMargin().bottom());
+#endif
+    if (sizeWithoutShadow.width() < 0) {
+        sizeWithoutShadow.setWidth(0);
+    }
+    if (sizeWithoutShadow.height() < 0) {
+        sizeWithoutShadow.setHeight(0);
+    }
+
+    QRect cursorRect = context_->cursorRectangleWrapper();
     QRect screenGeometry;
     // Try to apply the screen edge detection over the window, because if we
     // intent to use this with wayland. It we have no information above screen
     // edge.
     if (isWayland_) {
-        screenGeometry = QRect(QPoint(0, 0), window->size());
+        screenGeometry = window->frameGeometry();
         cursorRect.translate(window->framePosition());
         auto margins = window->frameMargins();
         cursorRect.translate(margins.left(), margins.top());
-        screenGeometry.translate(window->framePosition());
     } else {
         screenGeometry = window->screen()->geometry();
         auto pos = window->mapToGlobal(cursorRect.topLeft());
@@ -434,19 +440,20 @@ void fcitx::FcitxCandidateWindow::updateClientSideUI(
     }
 
     int x = cursorRect.left(), y = cursorRect.bottom();
-    if (cursorRect.left() + actualSize_.width() > screenGeometry.right()) {
-        x = screenGeometry.right() - actualSize_.width() + 1;
+    if (cursorRect.left() + sizeWithoutShadow.width() >
+        screenGeometry.right()) {
+        x = screenGeometry.right() - sizeWithoutShadow.width() + 1;
     }
 
     if (x < screenGeometry.left()) {
         x = screenGeometry.left();
     }
 
-    if (y + actualSize_.height() > screenGeometry.bottom()) {
+    if (y + sizeWithoutShadow.height() > screenGeometry.bottom()) {
         if (y > screenGeometry.bottom()) {
-            y = screenGeometry.bottom() - actualSize_.height() - 40;
+            y = screenGeometry.bottom() - sizeWithoutShadow.height() - 40;
         } else { /* better position the window */
-            y = y - actualSize_.height() -
+            y = y - sizeWithoutShadow.height() -
                 ((cursorRect.height() == 0) ? 40 : cursorRect.height());
         }
     }
@@ -459,6 +466,8 @@ void fcitx::FcitxCandidateWindow::updateClientSideUI(
     resize(actualSize_);
     // hide();
     QPoint newPosition(x, y);
+    newPosition -=
+        QPoint(theme_->shadowMargin().left(), theme_->shadowMargin().top());
     if (newPosition != position()) {
         if (isVisible()) {
             hide();
@@ -469,42 +478,43 @@ void fcitx::FcitxCandidateWindow::updateClientSideUI(
     show();
 }
 
-} // namespace fcitx
-
-void fcitx::FcitxCandidateWindow::mouseMoveEvent(QMouseEvent *event) {
+void FcitxCandidateWindow::mouseMoveEvent(QMouseEvent *event) {
     bool needRepaint = false;
+
+    bool prevHovered = false;
+    bool nextHovered = false;
     auto oldHighlight = highlight();
     hoverIndex_ = -1;
-    for (int idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
-        if (candidateRegions_[idx].contains(event->pos())) {
-            hoverIndex_ = idx;
-            break;
+
+    prevHovered = prevRegion_.contains(event->pos());
+    if (!prevHovered) {
+        nextHovered = nextRegion_.contains(event->pos());
+        if (!nextHovered) {
+            for (int idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
+                if (candidateRegions_[idx].contains(event->pos())) {
+                    hoverIndex_ = idx;
+                    break;
+                }
+            }
         }
     }
 
-    needRepaint = needRepaint || oldHighlight != highlight();
-
-    auto prevHovered = prevRegion_.contains(event->pos());
-    auto nextHovered = nextRegion_.contains(event->pos());
     needRepaint = needRepaint || prevHovered_ != prevHovered;
-    needRepaint = needRepaint || nextHovered_ != nextHovered;
     prevHovered_ = prevHovered;
+
+    needRepaint = needRepaint || nextHovered_ != nextHovered;
     nextHovered_ = nextHovered;
+
+    needRepaint = needRepaint || oldHighlight != highlight();
     if (needRepaint) {
         renderNow();
     }
 }
 
-void fcitx::FcitxCandidateWindow::mouseReleaseEvent(QMouseEvent *event) {
+void FcitxCandidateWindow::mouseReleaseEvent(QMouseEvent *event) {
     if (event->button() != Qt::LeftButton) {
         return;
     }
-    for (int idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
-        if (candidateRegions_[idx].contains(event->pos())) {
-            Q_EMIT candidateSelected(idx);
-            return;
-        }
-    }
 
     if (prevRegion_.contains(event->pos())) {
         Q_EMIT prevClicked();
@@ -513,10 +523,18 @@ void fcitx::FcitxCandidateWindow::mouseReleaseEvent(QMouseEvent *event) {
 
     if (nextRegion_.contains(event->pos())) {
         Q_EMIT nextClicked();
+        return;
+    }
+
+    for (int idx = 0, e = candidateRegions_.size(); idx < e; idx++) {
+        if (candidateRegions_[idx].contains(event->pos())) {
+            Q_EMIT candidateSelected(idx);
+            return;
+        }
     }
 }
 
-QSize fcitx::FcitxCandidateWindow::sizeHint() {
+QSize FcitxCandidateWindow::sizeHint() {
     auto minH =
         theme_->fontMetrics().ascent() + theme_->fontMetrics().descent();
 
@@ -532,16 +550,16 @@ QSize fcitx::FcitxCandidateWindow::sizeHint() {
     auto extraH = textMargin.top() + textMargin.bottom();
     if (!upperLayout_.text().isEmpty()) {
         auto size = upperLayout_.boundingRect();
-        height += std::max(minH, qCeil(size.height())) + extraH;
+        height += minH + extraH;
         updateIfLarger(width, size.width() + extraW);
     }
     if (!lowerLayout_.text().isEmpty()) {
         auto size = lowerLayout_.boundingRect();
-        height += std::max(minH, qCeil(size.height())) + extraH;
+        height += minH + extraH;
         updateIfLarger(width, size.width() + extraW);
     }
 
-    bool vertical = false;
+    bool vertical = theme_->vertical();
     if (layoutHint_ == FcitxCandidateLayoutHint::Vertical) {
         vertical = true;
     } else if (layoutHint_ == FcitxCandidateLayoutHint::Horizontal) {
@@ -590,7 +608,7 @@ QSize fcitx::FcitxCandidateWindow::sizeHint() {
     return {static_cast<int>(width), static_cast<int>(height)};
 }
 
-void fcitx::FcitxCandidateWindow::wheelEvent(QWheelEvent *event) {
+void FcitxCandidateWindow::wheelEvent(QWheelEvent *event) {
     if (!theme_ || !theme_->wheelForPaging()) {
         return;
     }
@@ -605,3 +623,5 @@ void fcitx::FcitxCandidateWindow::wheelEvent(QWheelEvent *event) {
         Q_EMIT nextClicked();
     }
 }
+
+} // namespace fcitx
