@@ -202,7 +202,10 @@ void QFcitxPlatformInputContext::invokeAction(QInputMethod::Action imAction,
     }
     if (FcitxQtInputContextProxy *proxy = validIC();
         proxy->supportInvokeAction()) {
-        proxy->invokeAction(action, cursorPosition);
+        if (cursorPosition >= 0 && cursorPosition <= preedit_.length()) {
+            auto ucs4Cursor = preedit_.left(cursorPosition).toUcs4().length();
+            proxy->invokeAction(action, ucs4Cursor);
+        }
     } else {
         if (cursorPosition <= 0 || cursorPosition >= preedit_.length()) {
             // qDebug() << action << cursorPosition;
@@ -259,6 +262,14 @@ void QFcitxPlatformInputContext::update(Qt::InputMethodQueries queries) {
         cursorRectChanged();
     }
 
+    if (queries & Qt::ImEnabled) {
+        if (!inputMethodAccepted() || !objectAcceptsInputMethod()) {
+            addCapability(data, FcitxCapabilityFlag_Disable);
+        } else {
+            removeCapability(data, FcitxCapabilityFlag_Disable);
+        }
+    }
+
     if (queries & Qt::ImHints) {
         Qt::InputMethodHints hints =
             Qt::InputMethodHints(query.value(Qt::ImHints).toUInt());
@@ -366,7 +377,7 @@ void QFcitxPlatformInputContext::setFocusObject(QObject *object) {
             createICData(window);
         }
     }
-    if (!window || (!inputMethodAccepted() && !objectAcceptsInputMethod())) {
+    if (!window) {
         lastWindow_ = nullptr;
         lastObject_ = nullptr;
         return;
@@ -376,17 +387,21 @@ void QFcitxPlatformInputContext::setFocusObject(QObject *object) {
         proxy->focusIn();
         // We need to delegate this otherwise it may cause self-recursion in
         // certain application like libreoffice.
-        QMetaObject::invokeMethod(this, "updateCursorRect",
-                                  Qt::QueuedConnection,
-                                  Q_ARG(QPointer<QWindow>, lastWindow_));
+        QMetaObject::invokeMethod(
+            this,
+            [this, window = QPointer<QWindow>(lastWindow_)]() {
+                if (window != lastWindow_) {
+                    return;
+                }
+                update(Qt::ImHints | Qt::ImEnabled);
+                updateCursorRect();
+            },
+            Qt::QueuedConnection);
     }
 }
 
-void QFcitxPlatformInputContext::updateCursorRect(QPointer<QWindow> window) {
-    if (window != lastWindow_) {
-        return;
-    }
-    if (validICByWindow(window.data())) {
+void QFcitxPlatformInputContext::updateCursorRect() {
+    if (validICByWindow(lastWindow_.data())) {
         cursorRectChanged();
     }
 }
@@ -454,8 +469,7 @@ void QFcitxPlatformInputContext::createInputContextFinished(
     if (proxy->isValid()) {
         QWindow *window = qApp->focusWindow();
         setFocusGroupForX11(uuid);
-        if (window && window == w && inputMethodAccepted() &&
-            objectAcceptsInputMethod()) {
+        if (window && window == w) {
             cursorRectChanged();
             proxy->focusIn();
         }
@@ -479,13 +493,20 @@ void QFcitxPlatformInputContext::createInputContextFinished(
     }
     flag |= FcitxCapabilityFlag_ClientSideInputPanel;
 
+    if (!inputMethodAccepted() || !objectAcceptsInputMethod()) {
+        flag |= FcitxCapabilityFlag_Disable;
+    }
+
+    // Notify fcitx of the effective bits from 0bit to 40bit
+    // (FcitxCapabilityFlag_Disable)
+    data->proxy->setSupportedCapability(0x1ffffffffffull);
+
     addCapability(*data, flag, true);
 }
 
 void QFcitxPlatformInputContext::updateCapability(const FcitxQtICData &data) {
     if (!data.proxy || !data.proxy->isValid())
         return;
-
     QDBusPendingReply<void> result = data.proxy->setCapability(data.capability);
 }
 
@@ -833,7 +854,7 @@ bool QFcitxPlatformInputContext::filterEvent(const QEvent *event) {
             }
         }
 
-        update(Qt::ImHints);
+        update(Qt::ImHints | Qt::ImEnabled);
         proxy->focusIn();
 
         auto stateToFcitx = state;
